  /*  dev.c

    helper functions for reading and processing dev namespace files (/dev)

    Copyright (C) 2001  Mike Sullivan

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Mike Sullivan may be reached by email at  sullivam@us.ibm.com

*/

/*
    Changes.

*/
#include "scsiname.h"

char *dev_recursive_find_kdev(int kdev, char *current_dirname);

/* 
 * dev_find_device - given a kdev (major and minor numbers) find /dev node
 *  arguments:
 *    kdev         kdev of node to locate
 *    devname      location to save name of matching device node
 *
 *  returns:
 *    0            OK
 *    1            Otherwise
*/
int dev_find_device(int kdev, char *devname)
{
	char *pathname=0;

	pathname=dev_recursive_find_kdev(kdev, DEVROOT);
	//TRACE (LOG_INFO, "dev::dev_find_device, return from dev_recursive_find_kdev pathname=%p\n",pathname);
	if (pathname) {
		strcpy(devname,pathname);
		free(pathname);
		return 0;
	}
	return 1;
}

/*
 * dev_recursive_find: helper routine to read the /dev namespace searching for a node
 *         that has a matching kdev
 *     kdev            - device number
 *     current_dirname - starting  directory location to search from
 * returns:
 *     the pathname of the /dev node that matches the kdev
 */
char *dev_recursive_find_kdev(int kdev, char *current_dirname) {
	struct dirent *dirp=NULL;
	DIR * dp=NULL;
	struct stat statbuf;
	char *pathname;
	char *parmdir=0;
	
	//TRACE (LOG_INFO, "dev::dev_recursive_find_kdev, opening dir %s\n",current_dirname);
	if ( (dp=opendir(current_dirname)) != NULL) {
		while( (dirp=readdir(dp)) != NULL) {
			if ( (!strcmp(dirp->d_name,".")) || (!strcmp(dirp->d_name,"..")) ) continue;
			pathname=malloc(strlen(current_dirname)+strlen(dirp->d_name)+2);
			strcpy(pathname,current_dirname);
			strcat(pathname,"/");
			strcat(pathname,dirp->d_name);
			lstat(pathname, &statbuf);
			//TRACE (LOG_INFO, "dev::dev_recursive_find_kdev, checking file %s\n",pathname);
			if ( S_ISDIR(statbuf.st_mode) ) {
				/* if we found a directory, drill down */
				parmdir=dev_recursive_find_kdev(kdev, pathname);
				if (parmdir) break;
			} else {
				/* its a regular devname file, compare its kdev value */
				if (statbuf.st_rdev == kdev) {
					closedir(dp);
					TRACE (LOG_INFO, "Found %s with matching kdev %x\n",pathname,kdev);
					return pathname;
				}
			}
			free(pathname);
		}
		closedir(dp);
	} else {
		TRACE (LOG_INFO, "dev::dev_recursive_find_kdev, cannot open dir %s\n",current_dirname);
	}
	return parmdir;
}

