 /*  main.c

    Hotplug interface to scsi device naming handler

    Copyright (C) 2001  Mike Sullivan

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Mike Sullivan may be reached by email at  sullivam@us.ibm.com

*/

/*
    Changes.
    Initial version Mike Sullivan   3/27/2001

*/
#include "scsiname.h"

int trace_is_open=FALSE;
int no_log=FALSE;
char temp[512];
char tbuffer[0x80000];
int tbuffer_index=0;
FILE *trace_fp;
time_t timeval;
int listflag=0;
extern char *scsi_types[];

char *event_types[]= {
  "add",
  "remove",
  "list"};

int dump_boot_log();


comparefns keyfns[MAX_KEY_FUNCTIONS] = { {"e2label", string_compare},
					 {"test",    string_compare} };

int main (int argc, char *argv[]) {
	int host,target,bus,lun,part;
	int all=0;
	int invoke_error=1;

	if (argc == 7) {
		TRACE (LOG_INFO,"scsiname Event %s received for %s:%s:%s:%s %s\n",argv[1],argv[2],argv[3],argv[4],argv[5],argv[6]);
		invoke_error=0;
		
	}
	if ( (argc == 3) && (!strcmp(argv[2],"all")) ) {
		TRACE (LOG_INFO,"scsiname Event %s received for %s\n",argv[1],argv[2]);
		invoke_error=0;
	}
	if ( (argc == 2) && !strcmp(argv[1],"list") ) {
		TRACE (LOG_INFO,"scsiname Event %s received\n",argv[1]);
		invoke_error=0;
	}

	if (invoke_error) {
		fprintf(stderr, "Invocation error. argc=%d Usage:\n",argc);
		fprintf(stderr, "\tscsiname <add/remove/list> host bus target lun part\n");
		return 0;
	}

	if (!strcmp(argv[1],event_types[2])) {
		listflag=1;
		main_process_sg_entry(-1,-1,-1,-1,-1,1);
		dump_boot_log();
		TRACE (LOG_INFO, "scsiname Event %s processing complete\n\n",argv[1]);
		return 0;
	}
	
	
	/* Read in user configuration info */
	if (config_read_info()) {
		dump_boot_log();
		return 0;
	}


	if (!strcmp(argv[2],"all")) {
		all=1;
		host=bus=target=lun=part=-1;
	} else {
		host=atoi(argv[2]);
		bus=atoi(argv[3]);
		target=atoi(argv[4]);
		lun=atoi(argv[5]);
		part=atoi(argv[6]);
	}
	
	if ( !strcmp(argv[1], event_types[0]) ) {
		main_process_sg_entry(host,bus,target,lun,part,all);
	} else { /* remove TBD */
	}

	if (argc == 7) {
		TRACE (LOG_INFO, "scsiname Event %s for %s:%s:%s:%s %s processing complete\n\n",argv[1],argv[2],argv[3], argv[4], argv[5], argv[6]);
	}
	if (argc == 3) {
		TRACE (LOG_INFO, "scsiname Event %s for %s processing complete\n\n",argv[1],argv[2]);
	}
	dump_boot_log();

	return 0;
}


int main_process_sg_entry(int host, int bus, int target, int lun, int part, int all) {
	sg_entry *sge;
	sg_entry *old_sge;
	config_entry *centry=0;

	sge=sg_read_entry(host,bus,target,lun,part,all);
	if (!sge) {
		TRACE (LOG_INFO, "Cannot find any matching devices (ignoring event)\n");
		return 0;
	} 
	if (!listflag) {
		while (sge) {
			sg_entry *old_sge;
			/* Find a user configuration entry that matches the device we received the event on. */
			if (!config_use_sg_names) {
				centry=sg_match_config_entry(sge,config_entry_head);
			} else {
				centry=config_create_default_entry(sge);
			}
			/* Given a matched entry perform the specified action (ie. create a symlink, rename a node) */
			if (centry) {
				TRACE (LOG_INFO, "Logging Matching user specific entry:\n");
				config_print_entry(centry);
				action_config_entry(centry,sge);
			}
			/* free old sge resources */
			if (sge->vendor) free(sge->vendor);
			if (sge->model) free (sge->model);
			if (sge->rev) free (sge->rev);
			if (sge->name) free (sge->name);
			if (sge->uid) free(sge->uid);
			if (sge->e2label) free(sge->e2label);
			if (sge->uuid) free(sge->uuid);
			old_sge=sge;
			sge=sge->next;
			free(old_sge);
		}
	} else {
		while (sge) {
			fprintf(stdout, "Info for device %s\n",sge->name);
			fprintf(stdout, "\tmajor=%d minor=%d\n",sge->major,sge->minor);
			fprintf(stdout, "\thost=%d  bus=%d  target=%d lun=%d  part=%d\n",
				sge->host,sge->bus,sge->target,sge->lun,sge->part);
			fprintf(stdout,"\tscsi_type=0x%x is ",sge->scsi_type);
			if ((sge->scsi_type & 0x1f) <= 0x10) {
				fprintf(stdout,"(%s)\n",scsi_types[sge->scsi_type]);
			} else {
				fprintf(stdout,"(UNKNOWN)\n");
			}
			if (sge->vendor && strlen(sge->vendor)) {
				fprintf(stdout, "\tvendor=%s\n",sge->vendor);
			}
			if (sge->model && strlen(sge->model)) {
				fprintf(stdout, "\tmodel=%s\n",sge->model);
			}
			if (sge->rev && strlen(sge->rev)) {
				fprintf(stdout, "\trev=%s\n",sge->rev);
			}
			if (sge->name && strlen(sge->name)) {
				fprintf(stdout, "\tname=%s\n",sge->name);
			}
			if (sge->type && strlen(sge->type)) {
				fprintf(stdout, "\ttype=%s\n",sge->type);
			}
			if (sge->e2label && strlen(sge->e2label)) {
				fprintf(stdout, "\te2label=%s\n",sge->e2label);
			}
			if (sge->uuid && strlen(sge->uuid)) {
				fprintf(stdout, "\tuuid=%s\n",sge->uuid);
			}
			old_sge=sge;
			sge=sge->next;
			free(old_sge);
		}
	}
	return 0;
}

/* string_compare - Compare the keyword+values of the config and sg entries provided
 *  arguments:
 *    mcentry      Config keyword+value entry
 *    sge          Complete sge entry
 *
 *  returns:
 *    0            Ignored
 *    1            Entries match
 *   -1            Entries do not match
 */
int string_compare(misc_config_entry *mcentry, sg_entry *sge) {
	int error=0;
	char *str_value=0;
	regmatch_t mbuf;
	char  int_value[8];

	memset(int_value,0,sizeof(int_value));
	if (!strcmp(mcentry->keyname,"vendor")) {
		str_value=sge->vendor;
	} else if (!strcmp(mcentry->keyname, "model")) {
		str_value=sge->model;
	} else if (!strcmp(mcentry->keyname, "rev")) {
		str_value=sge->rev;
	} else if (!strcmp(mcentry->keyname, "name")) {
		str_value=sge->name;
	} else if (!strcmp(mcentry->keyname, "e2label")) {
		str_value=sge->e2label;
	} else if (!strcmp(mcentry->keyname, "uuid")) {
		str_value=sge->uuid;
	} else if (!strcmp(mcentry->keyname, "major")) {
		sprintf(int_value,"%d",sge->major);
	} else if (!strcmp(mcentry->keyname, "minor")) {
		sprintf(int_value,"%d",sge->minor);
	} else if (!strcmp(mcentry->keyname, "host")) {
		sprintf(int_value,"%d",sge->host);
	} else if (!strcmp(mcentry->keyname, "bus")) {
		sprintf(int_value,"%d",sge->bus);
	} else if (!strcmp(mcentry->keyname, "target")) {
		sprintf(int_value,"%d",sge->target);
	} else if (!strcmp(mcentry->keyname, "lun")) {
		sprintf(int_value,"%d",sge->lun);
	} else if (!strcmp(mcentry->keyname, "part")) {
		if (sge->part == -1) {
			str_value="unknown";
		} else {
			sprintf(int_value,"%d",sge->part);
		}
	}
	if (strlen(int_value)) str_value=int_value;

	if (str_value) {
		if (regexec (&mcentry->preg, str_value, 1, &mbuf, 0) == 0) {
			TRACE(LOG_INFO, "  Match. file %s values: scsi=%s and user=%s\n",mcentry->keyname,str_value,mcentry->valuestr);
			error=1;
		} else {
			TRACE(LOG_INFO, "  No Match. file %s values: scsi=%s and user=%s\n",mcentry->keyname,str_value,mcentry->valuestr);
			error=-1;
		}
	} 
	return error;
}


/* dump_boot_log - Utility routine to process trace entries that occur early (before trace log can be opened)
 *   arguments:
 *    event        devfsd provided event. On first UNREGISTER accumulated entries will be
 *                 written to TRACE_FILE
 */
int dump_boot_log() {
	int index=0;
	pid_t pid=0;
	int retried=0;

try_again:
	if (retried) {
		pid=fork();
		if (pid) {
			return 0;
		}
		fprintf(stderr, "scsiname::dump_boot_log, sleeping 10 before retry\n");
		sleep(10);
	}
	
	if (trace_is_open == FALSE) {
		if ( (trace_fp=fopen(TRACE_FILE, "a")) == NULL) {
			if (!retried) {
				retried=1;
				goto try_again;
			}
			if (errno != EROFS) {
				fprintf(stderr, "scsiname::dump_boot_log, error opening trace file %s\n",TRACE_FILE);
			} else {
				TRACE (LOG_INFO, "scsiname::dump_boot_log, EROFS error when trying to open %s\n",TRACE_FILE);
			} 
			return 0;
		} else {
			trace_is_open=TRUE;
			fclose(trace_fp);
			TRACE (LOG_INFO, "SCSINAME version %s\n",SCSINAME_VERSION);
			while(index < tbuffer_index) {
				TRACE (LOG_INFO, &tbuffer[index]);
				index+=strlen(&tbuffer[index]);
			}
			return 1;
		}
	}
	return 1;
}


#define EXT2_SUPER_MAGIC    0xEF53
struct ext2_super_block {
	u_char   s_dummy1[56];
	u_char   s_magic[2];
	u_char   s_dummy2[46];
	u_char   s_uuid[16];
	u_char   s_volume_name[16];
};
#define ext2magic(s)	((uint) s.s_magic[0] + (((uint) s.s_magic[1]) << 8))

/* get_e2label_uuid - Get the volume id from the specified device (ext2 only for now)
 * arguments:
 *    device       devfs filesystem entry representing volume to read
 *    label        On success, label info is returned in this variable
 *    uuid         On success, UUID info is returned in this variable
 *
 *  returns:
 *    0            Label Returned
 *    !0           Failure
 */
int get_e2label_uuid(int fd, char *label, char *uuid) {
	size_t namesize;
	struct ext2_super_block e2sb;
	int error=1;
	
	if (lseek(fd, 1024, SEEK_SET) == 1024
	    && read(fd, (char *) &e2sb, sizeof(e2sb)) == sizeof(e2sb)
	    && (ext2magic(e2sb) == EXT2_SUPER_MAGIC)) {
		memcpy(uuid, e2sb.s_uuid, sizeof(e2sb.s_uuid));
		namesize=sizeof(e2sb.s_volume_name);
		memcpy(label, e2sb.s_volume_name, namesize);
		error=0;
	}
	return error;
}
