/*  scsiname.h

    Header file for  scsiname

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/
#define SCSINAME_DEBUG
#define SCSINAME_VERSION "0.1.0"
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>
#include <pwd.h>
#include <grp.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <dirent.h>
#include <fcntl.h>
#include <syslog.h>
#include <signal.h>
#include <regex.h>
#include <errno.h>
#include <dlfcn.h>
#include <time.h>
#include <rpcsvc/ypclnt.h>
#include <rpcsvc/yp_prot.h>
#include <linux/kdev_t.h>
#include <linux/devfs_fs.h>
#include <linux/major.h>
#include <scsi/scsi.h>

#define FAKENODE "/fakenode"
#define CONFIG_FILE "/etc/scsiname.conf"
#define DEVROOT "/dev"
#define FSTAB_PATH "/etc/fstab"
#define STRING_LENGTH               2048 
#define MAX_PATH_LENGTH             512
#define MAX_UID_LENGTH              256
#define MAX_DEVNAME_LENGTH          256
#define MAX_DEVICETYPE_LENGTH       16
#define MAX_DEVICEVENDORNAME_LENGTH 256
#define MAX_SYMLINK_LENGTH          128
#define MAX_KEYNAME_LENGTH          128
#define MAX_VALUESTR_LENGTH         128
#define MAX_KEY_FUNCTIONS           2

#define TRACE_FILE  "/etc/reg.trc"
extern int trace_is_open;
extern int no_log;
extern char temp[];
extern char tbuffer[];
extern int tbuffer_index;
extern FILE *trace_fp;
extern time_t timeval;

typedef struct _sg_entry {
	struct _sg_entry *next;
	int       major;                       /* kernel assigned major # for this dev   */
	int       minor;                       /* kernel assigned minor # for this dev   */ 
	int       host;
	int       bus;
        int       target;
	int       lun;
	int       part;
	int       scsi_type;
	char     *vendor;
	char     *model;
	char     *rev;
	char      type[16];                    /* BLK or CHR                             */
	char     *name;                        /* scsi device name (contains uid)        */
	char     *uid;                         /* derived uid from inquiries             */
	char     *e2label;                     /* ext 2 supported label                  */
	char     *uuid;                        /* uuid obtained from ext2 partition      */
} sg_entry;

typedef struct _misc_config_entry {
  struct _misc_config_entry *next;
  char      keyname[MAX_KEYNAME_LENGTH];
  char      valuestr[MAX_VALUESTR_LENGTH];
  regex_t   preg;
  int       (*matchfn) (struct _misc_config_entry *, sg_entry *);
} misc_config_entry;

typedef struct _specific_entry {
	//char      uid[MAX_UID_LENGTH];
	char      devname[MAX_DEVNAME_LENGTH];
	int       symlink_created;
	int       update_config;
	char      symlink[MAX_SYMLINK_LENGTH];
	char      alias[MAX_DEVNAME_LENGTH];
} specific_entry;

typedef struct _template_entry {
  char      template_symlink[MAX_SYMLINK_LENGTH];
} template_entry;

#define CONFIG_ENTRY_TYPE_SPECIFIC   1
#define CONFIG_ENTRY_TYPE_TEMPLATE   2
typedef struct _config_entry {
  struct    _config_entry *next;
  int       entrytype;
  union {
    specific_entry s;
    template_entry t;
  } type;
  misc_config_entry *misc_head;
} config_entry;

typedef struct _comparefns {
  char     *key;
  int       (*matchfn) (misc_config_entry *, sg_entry *);
} comparefns;

int e2label_compare(misc_config_entry *, sg_entry *);
int string_compare(misc_config_entry *, sg_entry *);

/*  Public data  */
extern struct _config_entry *config_entry_head;
extern struct _config_entry *template_entry_head;
extern comparefns keyfns[];
extern int config_use_sg_names;

/* main functions */
int main_process_sg_entry(int host, int bus, int target, int lun, int part, int all);
int get_e2label_uuid(int fd, char *label, char *uuid);

/* actions.c functions */
void action_config_entry(config_entry *centry, sg_entry *sge);

/* config.c functions */
int config_read_info(void);
void config_print_entry(config_entry *centry);
config_entry * config_create_default_entry(sg_entry *sge);

/* dev.c functions */
int dev_find_device(int kdev, char *devname);

/* sg.c functions */
sg_entry *sg_read_entry(int host, int target, int bus, int lun, int part, int all);
config_entry *sg_match_config_entry(sg_entry *sge, config_entry *centry);


// time(&timeval); 
// fprintf(trace_fp, "%s", ctime(&timeval)); 

#ifdef SCSINAME_DEBUG
#define TRACE(level, format, args...) \
    {if (trace_is_open) { \
       trace_fp=fopen(TRACE_FILE, "a+"); \
       fprintf(trace_fp, format, ##args);  \
       fclose(trace_fp); \
     } else { \
       if (no_log) { \
         fprintf(stderr, format, ##args); \
       } else { \
         sprintf(temp, format, ##args); \
         sprintf(&tbuffer[tbuffer_index],temp); \
         tbuffer_index+=strlen(temp);  \
       } \
     } \
    }

/*  Public data  */
extern int trace_is_open;  

#else
#define TRACE(level, format, args...) 
#endif



