/*  sg.c

    Main file for scsiname interaction with scsi subsystem through sg interface

    Copyright (C) 2001  Mike Sullivan

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Mike Sullivan may be reached by email at  sullivam@us.ibm.com
*/

/*
    Changes.
    Initial version Mike Sullivan   3/27/2001

*/
//#define DEV_DEBUG_TAPE
//#ifdef DEV_DEBUG_TAPE
//#define DEV_DEBUG
//#endif
//#define DEV_DEBUG_CDROM
//#ifdef DEV_DEBUG_CDROM
//#define DEV_DEBUG
//#endif

#include "scsiname.h"
char *scsi_types[]= {
  "DISK",
  "TAPE",
  "UNKNOWN",
  "PROCESSOR",
  "WORM",
  "ROM",
  "SCANNER",
  "MOD",
  "MEDIUM_CHANGER",
  "UNKNOWN",
  "UNKNOWN",
  "UNKNOWN",
  "UNKNOWN",
  "ENCLOSURE",
  "UNKNOWN",
  "UNKNOWN",
  "UNKNOWN"};

/* 
 * sg_print_entry - Display entry created from sg Inquiry
 *  arguments:
 *    sge        sg entry to display     
 */
void sg_print_entry(sg_entry *sge) {
	TRACE (LOG_INFO, "\tsg_entry ptr=%p\n",sge);
	TRACE (LOG_INFO, "\tmajor=%d minor=%d\n",sge->major,sge->minor);
	TRACE (LOG_INFO, "\thost=%d  bus=%d  target=%d lun=%d  part=%d\n",
	       sge->host,sge->bus,sge->target,sge->lun,sge->part);
	TRACE (LOG_INFO, "\tscsi_type=0x%x is ",sge->scsi_type);
	if ((sge->scsi_type & 0x1f) <= 0x10) {
		TRACE (LOG_INFO, "(%s)\n",scsi_types[sge->scsi_type]);
	} else {
		TRACE (LOG_INFO, "(UNKNOWN)\n");
	}
	if (sge->vendor && strlen(sge->vendor)) {
		TRACE (LOG_INFO, "\tvendor=%s\n",sge->vendor);
	}
	if (sge->model && strlen(sge->model)) {
		TRACE (LOG_INFO, "\tmodel=%s\n",sge->model);
	}
	if (sge->rev && strlen(sge->rev)) {
		TRACE (LOG_INFO, "\trev=%s\n",sge->rev);
	}
	if (sge->name && strlen(sge->name)) {
		TRACE (LOG_INFO, "\tname=%s\n",sge->name);
	}
	if (sge->uid && strlen(sge->uid)) {
		TRACE (LOG_INFO, "\tuid=%s\n",sge->uid);
	}
	if (sge->type && strlen(sge->type)) {
		TRACE (LOG_INFO, "\ttype=%s\n",sge->type);
	}
	if (sge->e2label && strlen(sge->e2label)) {
		TRACE (LOG_INFO, "\te2label=%s\n",sge->e2label);
	}
	if (sge->uuid && strlen(sge->uuid)) {
		TRACE (LOG_INFO, "\tuuid=%s\n",sge->uuid);
	}
	TRACE (LOG_INFO, "\tnext=%p\n",sge->next);
}

char *sg_getstr(char *scsi_cmd, int start, int max_lgth) {
	char *buf;
	int i,lgth;

	for(i=start+max_lgth-1; i>=0 && scsi_cmd[i] == ' '; i--) scsi_cmd[i]=0;

	lgth=strlen(&scsi_cmd[start]);
	if (lgth > max_lgth) lgth=max_lgth;

	buf=malloc(lgth+1);
	strncpy(buf,&scsi_cmd[start],lgth);
	buf[lgth]=0;
	
	return buf;
}

int sg_inquiry_standard(int fd, sg_entry *sge) {
	int error=0;
	unsigned char *scsi_cmd;
	unsigned char buf[1024];

	memset(buf,0, sizeof(buf));
	((int *)buf)[1]=sizeof(buf)-16;
	scsi_cmd=&buf[8];
	scsi_cmd[0]=INQUIRY;
	scsi_cmd[1]=sge->lun << 5;
	scsi_cmd[4]=255;

	error=ioctl(fd,1,buf);
	if (!error) {
		sge->vendor=sg_getstr(scsi_cmd, 8, 8);
		sge->model=sg_getstr(scsi_cmd,16,16);
		sge->rev=sg_getstr(scsi_cmd,32,4);
		sge->scsi_type=scsi_cmd[0]&0x1f;
	} else {
		TRACE (LOG_INFO, "sg::sg_inquiry_standard, ioctl failed %x errno 0x%x\n",error,errno);
	}
	return error;
}

char *sg_inquiry_evpd(int fd, sg_entry *sge) {
	int error=0;
	unsigned char *scsi_cmd;
	unsigned char buf[1024];
	unsigned char *evpd_page=NULL;

	memset(buf,0,sizeof(buf));
	((int *)buf)[1]=sizeof(buf)-16;
	scsi_cmd=&buf[8];
	scsi_cmd[0]=INQUIRY;
	scsi_cmd[1]=(sge->lun << 5) | 0x01;
	scsi_cmd[4]=255;

	error=ioctl(fd,1,buf);
	if (!error) {
		evpd_page=malloc(scsi_cmd[3]+4);
		if (evpd_page) 
			memcpy(evpd_page,scsi_cmd,scsi_cmd[3]+4);
	} else {
		TRACE (LOG_INFO, "sg::sg_inquiry_evpd, ioctl failed %x errno 0x%x\n",error,errno);
	}
	return evpd_page;
}

char *sg_inquiry_serialnum(int fd, sg_entry *sge) {
	int error=0;
	unsigned char *scsi_cmd;
	unsigned char buf[1024];
	unsigned char *serialnum=NULL;
	int i;
	
	memset(buf,0,sizeof(buf));
	((int *)buf)[1]=sizeof(buf)-16;
	scsi_cmd=&buf[8];
	scsi_cmd[0]=INQUIRY;
	scsi_cmd[1]=(sge->lun << 5) | 0x01;
	scsi_cmd[2]=0x80;
	scsi_cmd[4]=255;
	
	error=ioctl(fd,1,buf);
	if (!error) {
		serialnum=malloc(scsi_cmd[3]+1);
		if (serialnum) {
			memset(serialnum,0,scsi_cmd[3]+1);
			for(i=0; i<scsi_cmd[3]; i++) {
				if (scsi_cmd[4+i] > 0x20)
					memcpy(&serialnum[strlen(serialnum)],
					       &scsi_cmd[4+i],1);
			}
		} else {
			TRACE (LOG_INFO, "sg::sg_inquiry_serialnum, unable to malloc 0x%x bytes\n",scsi_cmd[3]+1);
		}
	} else {
		TRACE (LOG_INFO, "sg::sg_inquiry_serialnum, ioctl failed %x errno 0x%x\n",error,errno);
	}
	if (serialnum) {
		//TRACE (LOG_INFO, "sg::sg_inquiry_serialnum returns serialnum=%s\n",serialnum);
	}
	return serialnum;
}

char *sg_inquiry_deviceid(int fd, sg_entry *sge) {
	int error=0;
	unsigned char *scsi_cmd;
	unsigned char buf[1024];
	unsigned char *deviceid=NULL;
	int i,j;
	
	memset(buf,0,sizeof(buf));
	((int *)buf)[1]=sizeof(buf)-16;
	scsi_cmd=&buf[8];
	scsi_cmd[0]=INQUIRY;
	scsi_cmd[1]=(sge->lun << 5) | 0x01;
	scsi_cmd[2]=0x83;
	scsi_cmd[4]=255;
	
	error=ioctl(fd,1,buf);
	if (!error) {
		for(i=4; i<=scsi_cmd[3]+3; ) {
			if ( (scsi_cmd[i] & 0x0f) == 2) {  
				deviceid=malloc(scsi_cmd[i+3]+1);
				if (deviceid) {
					memset(deviceid,0,scsi_cmd[i+3]+1);
					for(j=0; j<scsi_cmd[i+3]; j++) {
						if (scsi_cmd[i+4+j] > 0x20)
							memcpy(&deviceid[strlen(deviceid)],&scsi_cmd[i+4+j],1);
					}
				} else {
					TRACE (LOG_INFO, "sg::sg_inquiry_deviceid cannot malloc 0x%x bytes\n",scsi_cmd[i+3]+1);
				}
				break;
			} else
				i+=scsi_cmd[i+3]+4;
		}
		if (!strlen(deviceid)) {
			for(i=4; i<=scsi_cmd[3]+3; ) {
				if ((scsi_cmd[i] & 0x0f) == 1) {
					deviceid=malloc(scsi_cmd[i+3]+1);
					if (deviceid) {
						memset(deviceid,0,scsi_cmd[i+3]+1);
						for(j=0; j<scsi_cmd[i+3]; j++)
							sprintf(&deviceid[strlen(deviceid)],"%02x",scsi_cmd[i+4+j]);
					} else {
						TRACE (LOG_INFO, "sg::sg_inquiry_deviceid cannot malloc 0x%x bytes\n",scsi_cmd[i+3]+1);
					}
					break;
				} else
					i+=deviceid[i+3]+4;
			}
		}
	}
	if (deviceid) {
		// TRACE (LOG_INFO, "sg::sg_inquiry_deviceid returns deviceid=%s\n",deviceid);
	}
	return deviceid;
}


int sg_load_inquiry_info(int fd, sg_entry *sge) {
	int error=0;
	unsigned char *evpd_page=0;
	int i;
	unsigned char *deviceid=NULL;
	unsigned char uid[512];
	
	sg_inquiry_standard(fd,sge);
	evpd_page=sg_inquiry_evpd(fd,sge);
	
	memset(uid,0,sizeof(uid));
	if (!evpd_page) {
		deviceid=sg_inquiry_serialnum(fd,sge);
		if (deviceid) { 
			goto unknown;
		} else {
			uid[0]='S';
			strcat(uid,sge->vendor);
			if (strlen(deviceid) >= (sizeof(uid)-strlen(uid)) )
				deviceid[sizeof(uid)-strlen(uid)]=0;
			strcat(uid,deviceid);
		}
	} else {
		for(i=4; i <= evpd_page[3]+3; i++) {
			if (evpd_page[i] == 0x83) {
				deviceid=sg_inquiry_deviceid(fd, sge);
				if (!deviceid) break;
				uid[0]='U';
				if (strlen(deviceid) >= (sizeof(uid)-strlen(uid)) )
					deviceid[sizeof(uid)-strlen(uid)]=0;
				strcat(uid,deviceid);
			}
		}
	}

	if (!strlen(uid)) {
		for(i=4; i <= evpd_page[3]+3; i++) {
			if (evpd_page[i] == 0x80) {
				deviceid=sg_inquiry_serialnum(fd,sge);
				if (!deviceid) break;
				uid[0]='S';
				strcat(uid,sge->vendor);
				if (strlen(deviceid) >= (sizeof(uid)-strlen(uid)) )
					deviceid[sizeof(uid)-strlen(uid)]=0;
				strcat(uid,deviceid);
			}
		}
	}
 unknown:
	if (!strlen(uid)) {
	        strcpy(uid,"Z");
	        if ( (sge->vendor) && (strlen(sge->vendor)) ) {
		        strcat(uid,sge->vendor);
		}
		if ( (sge->model) && (strlen(sge->model)) ) {
		        strcat(uid,sge->model);
		}
		if ( (sge->rev) && (strlen(sge->rev)) ) {
		        strcat(uid,sge->rev);
		}
	}
	sge->uid=malloc(strlen(uid)+1);
	if (sge->uid) {
		strcpy(sge->uid,uid);
	}
	if (evpd_page) free(evpd_page);
	if (deviceid)  free(deviceid);
	return error;
}

int sg_match_target(int major, int minor, int *host, int *bus, int *target, int *lun, int mode) {
	int fd=-1;
	int error=0;
	int info[2];

	error=mknod(FAKENODE, (0600 | mode),makedev(major, minor));
	if (error) {
		TRACE (LOG_INFO, "sg::sg_match_target, Error unable to make node for %x\n",makedev(major,minor));
	} else {
		
		fd=open(FAKENODE,O_RDONLY | O_NONBLOCK);
		unlink(FAKENODE);
		if (fd == -1) return fd;
		if (!(error=ioctl(fd,SCSI_IOCTL_GET_IDLUN, &info))) {
			if ( (*host == -1) &&
			     (*bus  == -1) &&
			     (*target == -1) &&
			     (*lun == -1) ) {
				*host=info[0]>>24;
				*bus=(info[0]>>16) & 0xff;
				*target=info[0] & 0xff;
				*lun=(info[0]>>8) & 0xff;
			} else {
				if ( (*host   != (info[0]>>24)) ||
				     (*bus    != ((info[0]>>16) & 0xff)) ||
				     (*target != (info[0] & 0xff)) ||
				     (*lun    != ((info[0]>>8) & 0xff)) ) {
					close(fd);
					fd=-1;
				}
			}
		}
	}
	return fd;
}

sg_entry *sg_clone(sg_entry *sge) {
	sg_entry *new_sge=NULL;
	
	new_sge=malloc(sizeof(sg_entry));
	if (new_sge) {
		memset(new_sge,0,sizeof(sg_entry));
		new_sge->major=sge->major;
		new_sge->minor=sge->minor;
		new_sge->host=sge->host;
		new_sge->bus=sge->bus;
		new_sge->target=sge->target;
		new_sge->lun=sge->lun;
		new_sge->part=sge->part;
		new_sge->scsi_type=sge->scsi_type;
		if (sge->vendor && strlen(sge->vendor))
			new_sge->vendor=strdup(sge->vendor);
		if (sge->model && strlen(sge->model))
			new_sge->model=strdup(sge->model);
		if (sge->rev && strlen(sge->rev))
			new_sge->rev=strdup(sge->rev);
		if (sge->uid && strlen(sge->uid))
			new_sge->uid=strdup(sge->uid);
	}
	return new_sge;
}

sg_entry *sg_create_cdrom(sg_entry *sge) {
	int minor,major;
	int host,bus,target,lun;
	int fd;
	sg_entry *cd_sge=NULL;

	if (!sge) return NULL;
	host=sge->host;
	bus=sge->bus;
	target=sge->target;
	lun=sge->lun;
	major=SCSI_CDROM_MAJOR;
	for(minor=0; minor<=255; minor++) {
		fd=sg_match_target(major,minor,&host,&bus,&target,&lun,S_IFBLK);
#ifndef DEV_DEBUG_CDROM
		if (fd == -1) continue;
		close(fd);
#endif
		cd_sge=sg_clone(sge);
		if (!cd_sge) return NULL;
		cd_sge->major=major;
		cd_sge->minor=minor;
		strcpy(cd_sge->type,"BLK");
		if (strlen(cd_sge->uid)) {
			cd_sge->name=malloc(strlen(cd_sge->uid)+1+sizeof("cdrom")+4);
			if (cd_sge->name) {
				sprintf(cd_sge->name,"%scdrom%d",cd_sge->uid,minor);
			}
		}
		break;
	}
	return cd_sge;
}

sg_entry *sg_create_tape(sg_entry *sge) {
	int minor,major;
	int host,bus,target,lun;
	int fd;
	sg_entry *tape_sge=NULL;
	sg_entry *head_sge=NULL;
	static char *formats[4]={"","l","m","a"};
	int mode;

	if (!sge) return NULL;
	host=sge->host;
	bus=sge->bus;
	target=sge->target;
	lun=sge->lun;
	major=SCSI_TAPE_MAJOR;
	for(minor=0; minor<=32; minor++) {
		fd=sg_match_target(major,minor,&host,&bus,&target,&lun,S_IFCHR);
#ifndef DEV_DEBUG_TAPE
		if (fd == -1) continue;
		close(fd);
#endif
		for(mode=0;mode<4;mode++) {
			tape_sge=sg_clone(sge);
			if (!tape_sge) return NULL;
			tape_sge->major=major;
			tape_sge->minor=minor+(mode<<5);
			strcpy(tape_sge->type,"CHR");
			if (strlen(tape_sge->uid)) {
				tape_sge->name=malloc(strlen(tape_sge->uid)+8);
				if (tape_sge->name) {
					sprintf(tape_sge->name,"%smt%s",tape_sge->uid,formats[mode]);
				}
			}
			tape_sge->next=head_sge;
			head_sge=tape_sge;
		}
		for(mode=0;mode<4;mode++) {
			tape_sge=sg_clone(sge);
			if (!tape_sge) return NULL;
			tape_sge->major=major;
			tape_sge->minor=minor+(mode<<5)+128;
			strcpy(tape_sge->type,"CHR");
			if (strlen(tape_sge->uid)) {
				tape_sge->name=malloc(strlen(tape_sge->uid)+8);
				if (tape_sge->name) {
					sprintf(tape_sge->name,"%smt%sn",tape_sge->uid,formats[mode]);
				}
			}
			tape_sge->next=head_sge;
			head_sge=tape_sge;
		}
		break;
	}
	return head_sge;
}

sg_entry *sg_create_disk(sg_entry *sge) {
	int minor,major;
	int host,bus,target,lun;
	int fd;
	sg_entry *head_sge=NULL;
	int found_disk=0;
	int error;

	if (!sge) return NULL;
	host=sge->host;
	bus=sge->bus;
	target=sge->target;
	lun=sge->lun;
	major=SCSI_DISK0_MAJOR;
	while(major <= SCSI_DISK7_MAJOR) {
		for(minor=0; minor<=255; minor=minor+16) {
			fd=sg_match_target(major,minor,&host,&bus,&target,&lun,S_IFBLK);
			if (fd == -1) continue;
			close(fd);
			sge=sg_clone(sge);
			if (!sge) return NULL;
			sge->major=major;
			sge->minor=minor;
			strcpy(sge->type,"BLK");
			if (strlen(sge->uid)) {
				sge->name=malloc(strlen(sge->uid)+1+sizeof("disc"));
				if (sge->name) {
					strcpy(sge->name,sge->uid);
					strcat(sge->name,"disc");
				}
			}
			sge->next=head_sge;
			head_sge=sge;
			found_disk=1;
			break;
		}
		if (!found_disk) {
			if (major == SCSI_DISK0_MAJOR) {
				major = SCSI_DISK1_MAJOR;
			} else {
				major++;
			}
		} else {
			break;
		}
			
	}
	if (found_disk) {
		for(minor=sge->minor+1;minor%16;minor++) {
			char label[256];
			char uuid[256];

			error=mknod(FAKENODE, 0600 | S_IFBLK, makedev(sge->major,minor));
			if (error) {
				TRACE (LOG_INFO, "Could NOT mkdev %s\n",FAKENODE);
				return NULL;
			}
			fd=open(FAKENODE,O_RDONLY | O_NONBLOCK);
			unlink(FAKENODE);
			if (fd == -1) continue;
			memset(label,0,256);
			memset(uuid,0,256);
			error=get_e2label_uuid(fd,label,uuid);
			close(fd);
			sge=sg_clone(sge);
			if (!sge) return NULL;
			if (!error) {
				if (strlen(label)) {
					sge->e2label=malloc(strlen(label)+1);
					if (sge->e2label) 
						strcpy(sge->e2label,label);
				}
				if (strlen(uuid)) {
					sge->uuid=malloc((strlen(uuid)*2)+1);
					if (sge->uuid) {
						int i;
						memset(sge->uuid,0,(strlen(uuid)*2)+1);
						for(i=0; i<strlen(uuid); i++) {
							sprintf(&sge->uuid[i*2],"%2x",uuid[i]&0xff);
						}
					}
				}
			}
			sge->major=major;
			sge->minor=minor;
			sge->part=minor%16;
			strcpy(sge->type,"BLK");
			if (strlen(sge->uid)) {
				sge->name=malloc(strlen(sge->uid)+1+sizeof("part")+4);
				if (sge->name) {
					sprintf(sge->name,"%spart%d",sge->uid,minor%16);
				}
			}
			sge->next=head_sge;
			head_sge=sge;
		}
	}
	return head_sge;
}
	
/* sg_read_entry - Verify and load info from sg for this device
 *    arguments:
 *    host       host number for the hotplugged device
 *    bus        bus number
 *    target     target number
 *    lun        lun number
 *
 *   returns:
 *    sg_entry*          entry built from sg kernel information
 *    0                  Otherwise
 */
sg_entry *sg_read_entry(int reqhost, int reqbus, int reqtarget, int reqlun, int reqpart, int all) {
	sg_entry *sge=NULL;
	sg_entry *head_sge=NULL;
	sg_entry *last_sge=NULL;
	int fd=0;
	struct stat statbuf;
	int error;
	int minor;
	int host,bus,target,lun;

	error=stat(FAKENODE, &statbuf);
	if (!error)
		unlink(FAKENODE);

	for(minor=0; minor<=255; minor++) {
	        host=reqhost;
		target=reqtarget;
		bus=reqbus;
		lun=reqlun;
#ifndef DEV_DEBUG
		fd=sg_match_target(SCSI_GENERIC_MAJOR,minor,&host,&bus,&target,&lun,S_IFCHR);
		if (fd == -1)  continue;
#endif
		sge=malloc(sizeof(sg_entry));
		if (!sge) return NULL;
		memset(sge,0,sizeof(sg_entry));
		sge->major=SCSI_GENERIC_MAJOR;
		sge->minor=minor;
		sge->part=-1;
		sge->host=host;
		sge->bus=bus;
		sge->target=target;
		sge->lun=lun;
		strcpy(sge->type,"CHR");
#ifndef DEV_DEBUG
		sg_load_inquiry_info(fd,sge);
#else
		sge->uid=malloc(80);
		strcpy(sge->uid,"fakeuid");
#ifdef DEV_DEBUG_TAPE
		sge->scsi_type=TYPE_TAPE;
#endif
#ifdef DEV_DEBUG_CDROM
		sge->scsi_type=TYPE_ROM;
#endif
#endif
		if (strlen(sge->uid)) {
			sge->name=malloc(strlen(sge->uid)+1+sizeof("generic"));
			if (sge->name) {
				strcpy(sge->name,sge->uid);
				strcat(sge->name,"generic");
			}
		}
		sge->next=head_sge;
		head_sge=sge;
		close(fd);

		/* now that generic device is found, create entries for real device types */
		switch(sge->scsi_type) {
		case TYPE_DISK:
		case TYPE_MOD:
			last_sge=sge=sg_create_disk(sge);
			while (last_sge) {
				if (last_sge->next) {
					last_sge=last_sge->next;
				} else {
					last_sge->next=head_sge;
					head_sge=sge;
					break;
				}
			}
			break;
		case TYPE_TAPE:
			last_sge=sge=sg_create_tape(sge);
			while (last_sge) {
				if (last_sge->next) {
					last_sge=last_sge->next;
				} else {
					last_sge->next=head_sge;
					head_sge=sge;
					break;
				}
			}
			break;
		case TYPE_ROM:
		case TYPE_WORM:
			sge=sg_create_cdrom(sge);
			if (sge) {
				sge->next=head_sge;
				head_sge=sge;
			}
			break;
		default:
			break;
		}
		if (!all) break;
	}
	
	sge=head_sge;
	while (sge) {
		TRACE (LOG_INFO, "Created scsi entry:\n");
		sg_print_entry(sge);
		sge=sge->next;
	} 
	return head_sge;
}


/* sg_match_config_entry - Given the sg entry corresponding to the device name
 *   find the corresponding user provided configuration entry.
 *   arguments:
 *   sge            Entry containing device information from sg Inquiry
 *   
 *   returns:
 *   config_entry*  If an matching entry was found
 *   0              Otherwise
 */
config_entry *sg_match_config_entry(sg_entry *sge, config_entry *centry) {
	misc_config_entry *mcentry;
	int match_count, best_count=0;
	config_entry *best_centry=0;
	int rc;
	
	while(centry) {
		if (centry->entrytype == CONFIG_ENTRY_TYPE_SPECIFIC) {
			TRACE (LOG_INFO, "Evaluating Specific Config Entry %p against sg Entry %p\n",centry,sge);
		} else {
			TRACE (LOG_INFO, "Evaluating Template Config Entry %p against sg Entry %p\n",centry,sge);
		}
		match_count=0;
		mcentry=centry->misc_head;
		while(mcentry) {
			rc=mcentry->matchfn(mcentry,sge);
			if (rc == 1)
				match_count++;
			else {
				if (rc == -1) {
					TRACE (LOG_INFO, "Config Entry %p invalidated\n",centry);
					goto new_centry;
				}
			}
			mcentry=mcentry->next;
		}
		TRACE (LOG_INFO, "Config Entry %p has matching value %x\n",centry,match_count);
		if (match_count > best_count) {
			best_count=match_count;
			best_centry=centry;
		}
new_centry:
		centry=centry->next;
	}
	if (best_centry) {
		TRACE (LOG_INFO, "Config Entry %p is best match %x\n",best_centry,best_count);  
	}
	return best_centry;
}
